@@ -42,6 +42,8 @@ namespace Util
         std::string getFilename(const size_t length);
     }
 
+    /// Hex to unsigned char
+    bool dataFromHexString(const std::string& hexString, std::vector<unsigned char>& data);
     /// Encode an integral ID into a string, with padding support.
     std::string encodeId(const unsigned number, const int padding = 5);
     /// Decode an integral ID from a string.
@@ -93,6 +95,9 @@ namespace Util
 #endif
     }
 
+    /// Returns the total physical memory (in kB) available in the system
+    size_t getTotalSystemMemory();
+
     /// Returns the process PSS in KB (works only when we have perms for /proc/pid/smaps).
     size_t getMemoryUsagePSS(const Poco::Process::PID pid);
 
@@ -103,6 +108,10 @@ namespace Util
     /// Example: "procmemstats: pid=123 rss=12400 pss=566"
     std::string getMemoryStats(FILE* file);
 
+    size_t getCpuUsage(const Poco::Process::PID pid);
+
+    size_t getStatFromPid(const Poco::Process::PID pid, int ind);
+
     std::string replace(std::string s, const std::string& a, const std::string& b);
 
     std::string formatLinesForLog(const std::string& s);
@@ -119,6 +128,9 @@ namespace Util
     /// Return a string that is unique across processes and calls.
     std::string UniqueId();
 
+    // Extract all json entries into a map.
+    std::map<std::string, std::string> JsonToMap(const std::string& jsonString);
+
     /// Trim spaces from the left. Just spaces.
     inline std::string& ltrim(std::string& s)
     {
@@ -203,14 +215,6 @@ namespace Util
         return trimmed(std::string(s));
     }
 
-    /// Check for the URI scheme validity.
-    /// For now just a basic sanity check, can be extended if necessary.
-    bool isValidURIScheme(const std::string& scheme);
-
-    /// Check for the URI host validity.
-    /// For now just a basic sanity check, can be extended if necessary.
-    bool isValidURIHost(const std::string& host);
-
     /// Given one or more patterns to allow, and one or more to deny,
     /// the match member will return true if, and only if, the subject
     /// matches the allowed list, but not the deny.
