@@ -37,6 +37,7 @@
 #include "Util.hpp"
 
 #include "common/FileUtil.hpp"
+#include "common/Seccomp.hpp"
 #include "common/SigUtil.hpp"
 #include "security.h"
 
@@ -110,6 +111,18 @@ public:
                     LOG_WRN("Cannot spawn " << tokens[1] << " children as requested.");
                 }
             }
+            else if (tokens.size() == 3 && tokens[0] == "setconfig")
+            {
+                // Currently only rlimit entries are supported.
+                if (!Rlimit::handleSetrlimitCommand(tokens))
+                {
+                    LOG_ERR("Unknown setconfig command: " << message);
+                }
+            }
+            else
+            {
+                LOG_ERR("Unknown command: " << message);
+            }
         }
         catch (const std::exception& exc)
         {
@@ -203,7 +216,7 @@ static void cleanupChildren()
         const auto it = childJails.find(exitedChildPid);
         if (it != childJails.end())
         {
-            LOG_INF("Child " << exitedChildPid << " has exited, removing its jail '" << it->second << "'.");
+            LOG_INF("Child " << exitedChildPid << " has exited, will remove its jail [" << it->second << "].");
             jails.emplace_back(it->second);
             childJails.erase(it);
         }
@@ -216,6 +229,7 @@ static void cleanupChildren()
     // Now delete the jails.
     for (const auto& path : jails)
     {
+        LOG_INF("Removing jail [" << path << "].");
         FileUtil::removeFile(path, true);
     }
 }
@@ -419,6 +433,21 @@ int main(int argc, char** argv)
             std::cout << "loolforkit version details: " << version << " - " << hash << std::endl;
             DisplayVersion = true;
         }
+        else if (std::strstr(cmd, "--rlimits") == cmd)
+        {
+            eq = std::strchr(cmd, '=');
+            const std::string rlimits = std::string(eq+1);
+            std::vector<std::string> tokens = LOOLProtocol::tokenize(rlimits, ';');
+            for (const std::string& cmdLimit : tokens)
+            {
+                const auto pair = LOOLProtocol::split(cmdLimit, ':');
+                std::vector<std::string> tokensLimit = { "setconfig", pair.first, pair.second };
+                if (!Rlimit::handleSetrlimitCommand(tokensLimit))
+                {
+                    LOG_ERR("Unknown rlimits command: " << cmdLimit);
+                }
+            }
+        }
 #if ENABLE_DEBUG
         // this process has various privileges - don't run arbitrary code.
         else if (std::strstr(cmd, "--unitlib=") == cmd)
@@ -474,7 +503,6 @@ int main(int argc, char** argv)
     if (!NoCapsForKit && !haveCorrectCapabilities())
     {
         std::cerr << "FATAL: Capabilities are not set for the loolforkit program." << std::endl;
-        std::cerr << "Please make sure that the current partition was *not* mounted with the 'nosuid' option." << std::endl;
         std::cerr << "If you are on SLES11, please set 'file_caps=1' as kernel boot option." << std::endl << std::endl;
         return Application::EXIT_SOFTWARE;
     }
