@@ -2,13 +2,21 @@
 * Control.Menubar
 */
 
-/* global $ _ map title vex revHistoryEnabled */
+/* global $ _ map vex revHistoryEnabled closebutton L */
 L.Control.Menubar = L.Control.extend({
 	// TODO: Some mechanism to stop the need to copy duplicate menus (eg. Help)
 	options: {
+		initial: [
+			{name: _('File'), disabled: true},
+			{name: _('Edit'), disabled: true},
+			{name: _('View'), disabled: true},
+			{name: _('Insert'), disabled: true},
+			{name: _('Tools'), disabled: true},
+		],
 		text:  [
-			{name: _('File'), type: 'menu', menu: [
-				{name: _('Save'), id: 'save', type: 'unocommand', uno: '.uno:Save'},
+			{name: _('File'), id: 'file', type: 'menu', menu: [
+				{name: _('Save'), id: 'save', type: 'action'},
+				{name: _('Save As'), id: 'saveas', type: 'action'},
 				{name: _('Print'), id: 'print', type: 'action'},
 				{name: _('See revision history'), id: 'rev-history', type: 'action'},
 				{name: _('Download as'), id: 'downloadas', type: 'menu', menu: [
@@ -18,26 +26,40 @@ L.Control.Menubar = L.Control.extend({
 					{name: _('Microsoft Word (.docx)'), id: 'downloadas-docx', type: 'action'}]}]
 			},
 			{name: _('Edit'), type: 'menu', menu: [
-				{name: _('Repair'), id: 'repair',  type: 'action'},
 				{name: _('Undo'), type: 'unocommand', uno: '.uno:Undo'},
 				{name: _('Redo'), type: 'unocommand', uno: '.uno:Redo'},
+				{name: _('Repair'), id: 'repair',  type: 'action'},
 				{type: 'separator'},
 				{name: _('Cut'), type: 'unocommand', uno: '.uno:Cut'},
 				{name: _('Copy'), type: 'unocommand', uno: '.uno:Copy'},
 				{name: _('Paste'), type: 'unocommand', uno: '.uno:Paste'},
+				{name: _('Select all'), type: 'unocommand', uno: '.uno:SelectAll'},
 				{type: 'separator'},
-				{name: _('Select all'), type: 'unocommand', uno: '.uno:SelectAll'}]
-			},
-			{name: _('View'), type: 'menu', menu: [
+				{name: _('Find & Replace'), id: 'findandreplace', type: 'action'},
+				{type: 'separator'},
+				{name: _('Track Changes'), type: 'menu', menu: [
+					{name: _('Record'), type: 'unocommand', uno: '.uno:TrackChanges'},
+					{name: _('Show'), type: 'unocommand', uno: '.uno:ShowTrackedChanges'},
+					{type: 'separator'},
+					{name: _('Accept All'), type: 'unocommand', uno: '.uno:AcceptAllTrackedChanges'},
+					{name: _('Reject All'), type: 'unocommand', uno: '.uno:RejectAllTrackedChanges'},
+					{name: _('Previous'), type: 'unocommand', uno: '.uno:PreviousTrackedChange'},
+					{name: _('Next'), type: 'unocommand', uno: '.uno:NextTrackedChange'}
+				]}
+			]},
+			{name: _('View'), id: 'view', type: 'menu', menu: [
 				{name: _('Full screen'), id: 'fullscreen', type: 'action'},
 				{type: 'separator'},
 				{name: _('Zoom in'), id: 'zoomin', type: 'action'},
 				{name: _('Zoom out'), id: 'zoomout', type: 'action'},
-				{name: _('Reset zoom'), id: 'zoomreset', type: 'action'}]
+				{name: _('Reset zoom'), id: 'zoomreset', type: 'action'},
+				{type: 'separator'},
+				{name: _('Formatting Marks'), type: 'unocommand', uno: '.uno:ControlCodes'},
+			]
 			},
 			{name: _('Insert'), type: 'menu', menu: [
 				{name: _('Image'), id: 'insertgraphic', type: 'action'},
-				{name: _('Comment'), type: 'unocommand', uno: '.uno:InsertAnnotation'},
+				{name: _('Comment...'), id: 'insertcomment', type: 'action'},
 				{type: 'separator'},
 				{name: _('Footnote'), type: 'unocommand', uno: '.uno:InsertFootnote'},
 				{name: _('Endnote'), type: 'unocommand', uno: '.uno:InsertEndnote'},
@@ -152,15 +174,26 @@ L.Control.Menubar = L.Control.extend({
 					{name: _('Cell'), type: 'unocommand', uno: '.uno:EntireCell'}]},
 					{name: _('Merge cells'), type: 'unocommand', uno: '.uno:MergeCells'}]
 			},
-			{name: _('Help'), type: 'menu', menu: [
+			{name: _('Tools'), id: 'tools', type: 'menu', menu: [
+				{name: _('Automatic spell checking'), type: 'unocommand', uno: '.uno:SpellOnline'},
+				{name: _('Language for selection'), type: 'menu', menu: [
+					{name: _('None (Do not check spelling)'), id: 'noneselection', type: 'unocommand', uno: '.uno:LanguageStatus?Language:string=Current_LANGUAGE_NONE'}]},
+				{name: _('Language for paragraph'), type: 'menu', menu: [
+					{name: _('None (Do not check spelling)'), id: 'noneparagraph', type: 'unocommand', uno: '.uno:LanguageStatus?Language:string=Paragraph_LANGUAGE_NONE'}]},
+				{name: _('Language for entire document'), type: 'menu', menu: [
+					{name: _('None (Do not check spelling)'), id: 'nonelanguage', type: 'unocommand', uno: '.uno:LanguageStatus?Language:string=Default_LANGUAGE_NONE'}]}
+			]},
+			{name: _('Help'), id: 'help', type: 'menu', menu: [
 				{name: _('Keyboard shortcuts'), id: 'keyboard-shortcuts', type: 'action'},
 				{name: _('About'), id: 'about', type: 'action'}]
-			}
+			},
+			{name: _('Close document'), id: 'closedocument', type: 'action'}
 		],
 
 		presentation: [
-			{name: _('File'), type: 'menu', menu: [
-				{name: _('Save'), id: 'save', type: 'unocommand', uno: '.uno:Save'},
+			{name: _('File'), id: 'file', type: 'menu', menu: [
+				{name: _('Save'), id: 'save', type: 'action'},
+				{name: _('Save As'), id: 'saveas', type: 'action'},
 				{name: _('Print'), id: 'print', type: 'action'},
 				{name: _('See revision history'), id: 'rev-history', type: 'action'},
 				{name: _('Download as'), id: 'downloadas', type: 'menu', menu: [
@@ -172,14 +205,16 @@ L.Control.Menubar = L.Control.extend({
 			{name: _('Edit'), type: 'menu', menu: [
 				{name: _('Undo'), type: 'unocommand', uno: '.uno:Undo'},
 				{name: _('Redo'), type: 'unocommand', uno: '.uno:Redo'},
+				{name: _('Repair'), id: 'repair',  type: 'action'},
 				{type: 'separator'},
 				{name: _('Cut'), type: 'unocommand', uno: '.uno:Cut'},
 				{name: _('Copy'), type: 'unocommand', uno: '.uno:Copy'},
 				{name: _('Paste'), type: 'unocommand', uno: '.uno:Paste'},
+				{name: _('Select all'), type: 'unocommand', uno: '.uno:SelectAll'},
 				{type: 'separator'},
-				{name: _('Select all'), type: 'unocommand', uno: '.uno:SelectAll'}]
+				{name: _('Find & Replace'), id: 'findandreplace', type: 'action'}]
 			},
-			{name: _('View'), type: 'menu', menu: [
+			{name: _('View'), id: 'view', type: 'menu', menu: [
 				{name: _('Full screen'), id: 'fullscreen', type: 'action'},
 				{type: 'separator'},
 				{name: _('Zoom in'), id: 'zoomin', type: 'action'},
@@ -188,6 +223,7 @@ L.Control.Menubar = L.Control.extend({
 			},
 			{name: _('Insert'), type: 'menu', menu: [
 				{name: _('Image'), id: 'insertgraphic', type: 'action'},
+				{name: _('Comment...'), id: 'insertcomment', type: 'action'},
 				{type: 'separator'},
 				{name: _('Special character...'), id: 'specialcharacter', type: 'action'}]
 			},
@@ -207,18 +243,25 @@ L.Control.Menubar = L.Control.extend({
 				{name: _('New slide'), id: 'insertpage', type: 'action'},
 				{name: _('Duplicate slide'), id: 'duplicatepage', type: 'action'},
 				{name: _('Delete slide'), id: 'deletepage', type: 'action'},
-				{type: 'separator'},
+				{type: 'separator', id: 'fullscreen-presentation-separator'},
 				{name: _('Fullscreen presentation'), id: 'fullscreen-presentation', type: 'action'}]
 			},
-			{name: _('Help'), type: 'menu', menu: [
+			{name: _('Tools'), id: 'tools', type: 'menu', menu: [
+				{name: _('Automatic spell checking'), type: 'unocommand', uno: '.uno:SpellOnline'},
+				{name: _('Language'), type: 'menu', menu: [
+					{name: _('None (Do not check spelling)'), id: 'nonelanguage', type: 'unocommand', uno: '.uno:LanguageStatus?Language:string=Default_LANGUAGE_NONE'}]}
+			]},
+			{name: _('Help'), id: 'help', type: 'menu', menu: [
 				{name: _('Keyboard shortcuts'), id: 'keyboard-shortcuts', type: 'action'},
 				{name: _('About'), id: 'about', type: 'action'}]
-			}
+			},
+			{name: _('Close document'), id: 'closedocument', type: 'action'}
 		],
 
 		spreadsheet: [
-			{name: _('File'), type: 'menu', menu: [
-				{name: _('Save'), id: 'save', type: 'unocommand', uno: '.uno:Save'},
+			{name: _('File'), id: 'file', type: 'menu', menu: [
+				{name: _('Save'), id: 'save', type: 'action'},
+				{name: _('Save As'), id: 'saveas', type: 'action'},
 				{name: _('Print'), id: 'print', type: 'action'},
 				{name: _('See revision history'), id: 'rev-history', type: 'action'},
 				{name: _('Download as'), id:'downloadas', type: 'menu', menu: [
@@ -230,19 +273,21 @@ L.Control.Menubar = L.Control.extend({
 			{name: _('Edit'), type: 'menu', menu: [
 				{name: _('Undo'), type: 'unocommand', uno: '.uno:Undo'},
 				{name: _('Redo'), type: 'unocommand', uno: '.uno:Redo'},
+				{name: _('Repair'), id: 'repair',  type: 'action'},
 				{type: 'separator'},
 				{name: _('Cut'), type: 'unocommand', uno: '.uno:Cut'},
 				{name: _('Copy'), type: 'unocommand', uno: '.uno:Copy'},
 				{name: _('Paste'), type: 'unocommand', uno: '.uno:Paste'},
+				{name: _('Select all'), type: 'unocommand', uno: '.uno:SelectAll'},
 				{type: 'separator'},
-				{name: _('Select all'), type: 'unocommand', uno: '.uno:SelectAll'}]
+				{name: _('Find & Replace'), id: 'findandreplace', type: 'action'}]
 			},
-			{name: _('View'), type: 'menu', menu: [
+			{name: _('View'), id: 'view', type: 'menu', menu: [
 				{name: _('Full screen'), id: 'fullscreen', type: 'action'}]
 			},
 			{name: _('Insert'), type: 'menu', menu: [
 				{name: _('Image'), id: 'insertgraphic', type: 'action'},
-				{name: _('Comment'), type: 'unocommand', uno: '.uno:InsertAnnotation'},
+				{name: _('Comment...'), id: 'insertcomment', type: 'action'},
 				{type: 'separator'},
 				{name: _('Row'), type: 'unocommand', uno: '.uno:InsertRows'},
 				{name: _('Column'), type: 'unocommand', uno: '.uno:InsertColumns'},
@@ -256,40 +301,107 @@ L.Control.Menubar = L.Control.extend({
 				{name: _('Delete row'), type: 'unocommand', uno: '.uno:DeleteRows'},
 				{name: _('Delete column'), type: 'unocommand', uno: '.uno:DeleteColumns'}]
 			},
-			{name: _('Help'), type: 'menu', menu: [
+			{name: _('Tools'), id: 'tools', type: 'menu', menu: [
+				{name: _('Automatic spell checking'), type: 'unocommand', uno: '.uno:SpellOnline'},
+				{name: _('Language'), type: 'menu', menu: [
+					{name: _('None (Do not check spelling)'), id: 'nonelanguage', type: 'unocommand', uno: '.uno:LanguageStatus?Language:string=Default_LANGUAGE_NONE'}]}
+			]},
+			{name: _('Help'), id: 'help', type: 'menu', menu: [
 				{name: _('Keyboard shortcuts'), id: 'keyboard-shortcuts', type: 'action'},
 				{name: _('About'), id: 'about', type: 'action'}]
-			}
+			},
+			{name: _('Close document'), id: 'closedocument', type: 'action'}
 		],
 
 		commandStates: {},
 
-		allowedViewModeActions: ['downloadas-pdf', 'downloadas-odt', 'downloadas-doc', 'downloadas-docx',
-								 'downloadas-odp', 'downloadas-ppt', 'downloadas-pptx',
-								 'downloadas-ods', 'downloadas-xls', 'downloadas-xlsx',
-								 'fullscreen', 'zoomin', 'zoomout', 'zoomreset',
-								 'about', 'keyboard-shortcuts']
+		// Only these menu options will be visible in readonly mode
+		allowedReadonlyMenus: ['file', 'downloadas', 'view', 'help'],
+
+		allowedViewModeActions: [
+			'downloadas-pdf', 'downloadas-odt', 'downloadas-doc', 'downloadas-docx', // file menu
+			'downloadas-odp', 'downloadas-ppt', 'downloadas-pptx', // file menu
+			'downloadas-ods', 'downloadas-xls', 'downloadas-xlsx', // file menu
+			'fullscreen', 'zoomin', 'zoomout', 'zoomreset', // view menu
+			'about', 'keyboard-shortcuts' // help menu
+		]
 	},
 
 	onAdd: function (map) {
 		this._initialized = false;
-		var docContainer = map.options.documentContainer;
-		this._menubarCont = L.DomUtil.create('ul', 'sm sm-simple', docContainer.parentElement);
-		this._menubarCont.id = 'main-menu';
+		this._menubarCont = L.DomUtil.get('main-menu');
+		this._initializeMenu(this.options.initial);
 
 		map.on('doclayerinit', this._onDocLayerInit, this);
-		map.on('commandstatechanged', this._onCommandStateChanged, this);
+		map.on('addmenu', this._addMenu, this);
+		map.on('commandvalues', this._onInitMenu, this);
 	},
 
-	_onCommandStateChanged: function(e) {
-		// Store information about enabled/disabled commands
-		// Used later just before showing menu to enable/disable menu items
-		if (e.state === 'enabled' || e.state === 'disabled') {
-			this.options.commandStates[e.commandName] = e.state;
+	_addMenu: function (e) {
+		var alreadyExists = L.DomUtil.get('menu-' + e.id);
+		if (alreadyExists)
+			return;
+
+		var liItem = L.DomUtil.create('li', '');
+		liItem.id = 'menu-' + e.id;
+		if (map._permission === 'readonly') {
+			L.DomUtil.addClass(liItem, 'readonly');
+		}
+		var aItem = L.DomUtil.create('a', '', liItem);
+		$(aItem).text(e.label);
+		$(aItem).data('id', e.id);
+		$(aItem).data('type', 'action');
+		$(aItem).data('postmessage', 'true');
+		this._menubarCont.insertBefore(liItem, this._menubarCont.firstChild);
+	},
+
+	_createLangMenuItem: function (lang, command) {
+		var liItem, aItem;
+		liItem = L.DomUtil.create('li', '');
+		aItem = L.DomUtil.create('a', '', liItem);
+		$(aItem).text(lang);
+		$(aItem).data('type', 'unocommand');
+		$(aItem).data('uno', '.uno:LanguageStatus?Language:string=' + command);
+		return liItem;
+	},
+
+	_onInitMenu: function (e) {
+		if (e.commandName === '.uno:LanguageStatus' && L.Util.isArray(e.commandValues)) {
+			var resetLang = _('Reset to Default Language'), translated, neutral;
+			var languages  = [];
+
+			e.commandValues.forEach(function(language) {
+				languages.push({translated: _(language), neutral: language});
+			});
+			languages.sort(function(a, b) {
+				return a.translated < b.translated ? -1 : a.translated > b.translated ? 1 : 0;
+			});
+
+			$menuSelection = $('#menu-noneselection').parent();
+			$menuParagraph = $('#menu-noneparagraph').parent();
+			$menuDefault = $('#menu-nonelanguage').parent();
+			for (var lang in languages) {
+				translated = languages[lang].translated;
+				neutral = languages[lang].neutral;
+				$menuSelection.append(this._createLangMenuItem(translated, encodeURIComponent('Current_' + neutral)));
+				$menuParagraph.append(this._createLangMenuItem(translated, encodeURIComponent('Paragraph_' + neutral)));
+				$menuDefault.append(this._createLangMenuItem(translated, encodeURIComponent('Default_' + neutral)));
+			}
+			$menuSelection.append(this._createMenu([{type: 'separator'}]));
+			$menuParagraph.append(this._createMenu([{type: 'separator'}]));
+			$menuDefault.append(this._createMenu([{type: 'separator'}]));
+			$menuSelection.append(this._createLangMenuItem(resetLang, 'Current_RESET_LANGUAGES'));
+			$menuParagraph.append(this._createLangMenuItem(resetLang, 'Paragraph_RESET_LANGUAGES'));
+			$menuDefault.append(this._createLangMenuItem(resetLang, 'Default_RESET_LANGUAGES'));
 		}
 	},
 
 	_onDocLayerInit: function() {
+		// clear initial menu
+		while (this._menubarCont.hasChildNodes()) {
+			this._menubarCont.removeChild(this._menubarCont.firstChild);
+		}
+
 		// Add document specific menu
 		var docType = this._map.getDocType();
 		if (docType === 'text') {
@@ -318,6 +430,28 @@ L.Control.Menubar = L.Control.extend({
 		$('#main-menu').bind('beforeshow.smapi', {self: this}, this._beforeShow);
 		$('#main-menu').bind('click.smapi', {self: this}, this._onClicked);
 
+		// SmartMenus mobile menu toggle button
+		$(function() {
+			var $mainMenuState = $('#main-menu-state');
+			if ($mainMenuState.length) {
+				// animate mobile menu
+				$mainMenuState.change(function(e) {
+					var $menu = $('#main-menu');
+					if (this.checked) {
+						$menu.hide().slideDown(250, function() { $menu.css('display', ''); });
+					} else {
+						$menu.show().slideUp(250, function() { $menu.css('display', ''); });
+					}
+				});
+				// hide mobile menu beforeunload
+				$(window).bind('beforeunload unload', function() {
+					if ($mainMenuState[0].checked) {
+						$mainMenuState[0].click();
+					}
+				});
+			}
+		});
+
 		this._initialized = true;
 	},
 
@@ -325,15 +459,16 @@ L.Control.Menubar = L.Control.extend({
 		if ($(menu).hasClass('highlighted')) {
 			$('#main-menu').smartmenus('menuHideAll');
 		}
+
+		var $mainMenuState = $('#main-menu-state');
+		if (!$(menu).hasClass('has-submenu') && $mainMenuState[0].checked) {
+			$mainMenuState[0].click();
+		}
 	},
 
 	_beforeShow: function(e, menu) {
 		var self = e.data.self;
 		var items = $(menu).children().children('a').not('.has-submenu');
-		var ua = window.navigator.userAgent;
-		var msie = ua.indexOf('MSIE '); // IE 10 or older
-		var trident = ua.indexOf('Trident/'); // IE 11
-		var edge = ua.indexOf('Edge/'); // Microsoft Edge
 		$(items).each(function() {
 			var aItem = this;
 			var type = $(aItem).data('type');
@@ -341,13 +476,28 @@ L.Control.Menubar = L.Control.extend({
 			if (map._permission === 'edit') {
 				if (type === 'unocommand') { // enable all depending on stored commandStates
 					var unoCommand = $(aItem).data('uno');
-					if (self.options.commandStates[unoCommand] === 'disabled') {
+					var itemState = map['stateChangeHandler'].getItemValue(unoCommand);
+					if (itemState === 'disabled') {
 						$(aItem).addClass('disabled');
 					} else {
 						$(aItem).removeClass('disabled');
 					}
+					if (unoCommand.indexOf('.uno:LanguageStatus') !== -1) {
+						var lang = map['stateChangeHandler'].getItemValue('.uno:LanguageStatus');
+						var data = decodeURIComponent($(aItem).data('uno'));
+						if (data.indexOf(lang) !== -1) {
+							$(aItem).addClass('lo-menu-item-checked');
+						} else {
+							$(aItem).removeClass('lo-menu-item-checked');
+						}
+					}
+					else if (itemState === 'true') {
+						$(aItem).addClass('lo-menu-item-checked');
+					} else {
+						$(aItem).removeClass('lo-menu-item-checked');
+					}
 				} else if (type === 'action') { // enable all except fullscreen on windows
-					if (id === 'fullscreen' && (msie > 0 || trident > 0 || edge > 0)) { // Full screen works weirdly on IE 11 and on Edge
+					if (id === 'fullscreen' && (L.Browser.ie || L.Browser.edge)) { // Full screen works weirdly on IE 11 and on Edge
 						$(aItem).addClass('disabled');
 						var index = self.options.allowedViewModeActions.indexOf('fullscreen');
 						if (index > 0) {
@@ -378,16 +528,148 @@ L.Control.Menubar = L.Control.extend({
 		});
 	},
 
-	_executeAction: function(id) {
-		if (id === 'print') {
+	_onClickFindAndReplace: function() {
+		var findReplaceContent =
+		    '\
+			<table class="findreplacetable">\
+				<tr>\
+					<td>\
+						<label for="findthis">' + _('Find') + '</label>\
+					</td>\
+					<td>\
+						<input id="findthis" name="findthis">\
+					</td>\
+				</tr>\
+				<tr>\
+					<td>\
+						<label for="replacewith">' + _('Replace with') + '</label>\
+					</td>\
+					<td>\
+						<input name="replacewith">\
+					</td>\
+				</tr>\
+			</table>\
+			';
+		var mouseMoveFunc;
+		vex.dialog.open({
+			showCloseButton: true,
+			escapeButtonCloses: true,
+			className: 'vex-theme-plain findReplaceVex',
+			message: _('Find & Replace'),
+			input: findReplaceContent,
+			buttons: [
+				$.extend({}, vex.dialog.buttons.replace, {
+					text: _('Replace'),
+					click: function($vexContent, e) {
+						$vexContent.data().vex.option = 'replace';
+					}}),
+				$.extend({}, vex.dialog.buttons.replaceAll, {
+					text: _('Replace All'),
+					click: function($vexContent, e) {
+						$vexContent.data().vex.option = 'replaceAll';
+					}}),
+				$.extend({}, vex.dialog.buttons.findPrev, {
+					text: _('Previous'),
+					className: 'btnArrow',
+					click: function($vexContent, e) {
+						$vexContent.data().vex.option = 'previous';
+					}}),
+				$.extend({}, vex.dialog.buttons.findNext, {
+					text: _('Next'),
+					className: 'btnArrow',
+					click: function($vexContent, e) {
+						$vexContent.data().vex.option = 'next';
+					}})
+			],
+			afterOpen: function(e) {
+				$('.vex-overlay').remove();
+				$('.vex').css('position', 'static');
+				var selected = null;
+				var xPos = 0, yPos = 0;
+				var xElem = 0, yElem = 0;
+				var maxH = window.innerHeight, maxW = window.innerWidth;
+
+				if (map.getDocType() === 'text') {
+					$('#findthis').on('input', function() {
+						if (this.value.length != 0) {
+							map.search(this.value, false, '', 0, true);
+						}
+					});
+				}
+				$('.vex-content').on('mousedown', function(e) {
+					selected = this;
+					selected.style.cursor = 'move';
+					xElem = xPos - selected.offsetLeft;
+					yElem = yPos - selected.offsetTop;
+				});
+				$('.vex-content').on('mouseup', function(e) {
+					selected.style.cursor = 'default';
+					selected = null;
+				});
+				var mouseMoveFunc = function(e) {
+					xPos = e.pageX;
+					yPos = e.pageY;
+					if (selected !== null) {
+						var isOutVert = (yPos - yElem >= 0 && (yPos - yElem + selected.offsetHeight) <= maxH);
+						var isOutHor = (xPos - xElem >= 0 && (xPos - xElem + selected.offsetWidth) <= maxW);
+						if (isOutHor) {
+							selected.style.left = (xPos - xElem) + 'px';
+						}
+						if (isOutVert) {
+							selected.style.top = (yPos - yElem) + 'px';
+						}
+					}
+				};
+				$(document).on('mousemove', mouseMoveFunc);
+			},
+			afterClose: function(e) {
+				$(document).off('mousemove', mouseMoveFunc);
+			},
+			onSubmit: function(event) {
+				var $vexContent = $(this).parent();
+				event.preventDefault();
+				event.stopPropagation();
+
+				var opt = $vexContent.data().vex.option;
+				var findText = this.findthis.value;
+				var replaceText = this.replacewith.value;
+
+				if (findText.length != 0) {
+					if (opt === 'next') {
+						map.search(findText);
+					}
+					else if (opt === 'previous') {
+						map.search(findText, true);
+					}
+					else if (opt === 'replace') {
+						map.search(findText, false, replaceText, 2);
+					}
+					else if (opt === 'replaceAll') {
+						map.search(findText, false, replaceText, 3);
+					}
+				}
+			}
+		}, this);
+	},
+
+	_executeAction: function(item) {
+		var id = $(item).data('id');
+		if (id === 'save') {
+			map.save(true, true);
+		} else if (id === 'saveas') {
+			map.fire('postMessage', {msgId: 'UI_SaveAs'});
+		} else if (id === 'print') {
 			map.print();
 		} else if (id.startsWith('downloadas-')) {
 			var format = id.substring('downloadas-'.length);
-			// remove the extension if any
-			var fileName = title.substr(0, title.lastIndexOf('.')) || title;
-			// check if it is empty
+			var fileName = map['wopi'].BaseFileName;
+			fileName = fileName.substr(0, fileName.lastIndexOf('.'));
 			fileName = fileName === '' ? 'document' : fileName;
 			map.downloadAs(fileName + '.' + format, format);
+		} else if (id === 'findandreplace') {
+			this._onClickFindAndReplace();
+		} else if (id === 'insertcomment') {
+			map.insertComment();
 		} else if (id === 'insertgraphic') {
 			L.DomUtil.get('insertgraphic').click();
 		} else if (id === 'specialcharacter') {
@@ -441,8 +723,14 @@ L.Control.Menubar = L.Control.extend({
 		} else if (id === 'rev-history') {
 			// if we are being loaded inside an iframe, ask
 			// our host to show revision history mode
-			map.fire('postMessage', {msgId: 'rev-history'});
-		} else if (id === 'repair') {
+			map.fire('postMessage', {msgId: 'rev-history', args: {Deprecated: true}});
+			map.fire('postMessage', {msgId: 'UI_FileVersions'});
+		} else if (id === 'closedocument') {
+			map.fire('postMessage', {msgId: 'close', args: {EverModified: map._everModified, Deprecated: true}});
+			map.fire('postMessage', {msgId: 'UI_Close', args: {EverModified: map._everModified}});
+			map.remove();
+		}
+		else if (id === 'repair') {
 			map._socket.sendMessage('commandvalues command=.uno:DocumentRepair');
 		} else if (id === 'a4portrait') {
 			map.sendUnoCommand('.uno:AttributePageSize {"AttributePageSize.Width":{"type":"long", "value": "21000"},"AttributePageSize.Height":{"type":"long", "value": "29700"}}');
@@ -469,6 +757,11 @@ L.Control.Menubar = L.Control.extend({
 			map.sendUnoCommand('.uno:AttributePageSize {"AttributePageSize.Height":{"type":"long", "value": "21590"},"AttributePageSize.Width":{"type":"long", "value": "35560"}}');
 			map.sendUnoCommand('.uno:AttributePage {"AttributePage.Landscape":{"type":"boolean", "value": "true"}}');
 		}
+
+		// Inform the host if asked
+		if ($(item).data('postmessage') === 'true') {
+			map.fire('postMessage', {msgId: 'Clicked_Button', args: {Id: id} });
+		}
 	},
 
 	_onDeleteSlide: function(e) {
@@ -484,27 +777,51 @@ L.Control.Menubar = L.Control.extend({
 			var unoCommand = $(item).data('uno');
 			map.sendUnoCommand(unoCommand);
 		} else if (type === 'action') {
-			var id = $(item).data('id');
-			self._executeAction(id);
+			self._executeAction(item);
 		}
-		map.focus();
+
+		if ($(item).data('id') !== 'insertcomment')
+			map.focus();
 	},
 
 	_createMenu: function(menu) {
 		var itemList = [];
 		for (var i in menu) {
-			if (menu[i].type === 'action' &&
-			    menu[i].id === 'rev-history' &&
-			    !revHistoryEnabled) {
+			if (menu[i].id === 'about' && (L.DomUtil.get('about-dialog') === null)) {
 				continue;
 			}
 
+			if (map._permission === 'readonly' && menu[i].type === 'menu') {
+				var found = false;
+				for (var j in this.options.allowedReadonlyMenus) {
+					if (this.options.allowedReadonlyMenus[j] === menu[i].id) {
+						found = true;
+						break;
+					}
+				}
+				if (!found)
+					continue;
+			}
+
+			if (menu[i].type === 'action') {
+				if ((menu[i].id === 'rev-history' && !revHistoryEnabled) ||
+					(menu[i].id === 'closedocument' && !closebutton)) {
+					continue;
+				}
+			}
+
 			if (menu[i].id === 'print' && this._map['wopi'].HidePrintOption)
 				continue;
 
 			if (menu[i].id === 'save' && this._map['wopi'].HideSaveOption)
 				continue;
 
+			if (menu[i].id === 'saveas' && this._map['wopi'].UserCanNotWriteRelative)
+				continue;
+
+			if (menu[i].id && menu[i].id.startsWith('fullscreen-presentation') && this._map['wopi'].HideExportOption)
+				continue;
+
 			// Keep track of all 'downloadas-' options and register them as
 			// export formats with docLayer which can then be publicly accessed unlike
 			// this Menubar control for which there doesn't seem to be any easy way
@@ -518,7 +835,14 @@ L.Control.Menubar = L.Control.extend({
 			}
 
 			var liItem = L.DomUtil.create('li', '');
-			var aItem = L.DomUtil.create('a', '', liItem);
+			if (menu[i].id) {
+				liItem.id = 'menu-' + menu[i].id;
+				if (menu[i].id === 'closedocument' && map._permission === 'readonly') {
+					// see corresponding css rule for readonly class usage
+					L.DomUtil.addClass(liItem, 'readonly');
+				}
+			}
+			var aItem = L.DomUtil.create('a', menu[i].disabled ? 'disabled' : '', liItem);
 			aItem.innerHTML = menu[i].name;
 
 			if (menu[i].type === 'menu') {
