@@ -142,6 +142,11 @@ L.TileLayer = L.GridLayer.extend({
 			'tiletwipwidth=' + this.options.tileWidthTwips + ' ' +
 			'tiletwipheight=' + this.options.tileHeightTwips;
 
+		this._followThis = -1;
+		this._editorId = -1;
+		this._followUser = false;
+		this._followEditor = false;
+
 		// Mark visible area as dirty by default.
 		this._invalidateClientVisibleArea();
 	},
@@ -171,6 +176,7 @@ L.TileLayer = L.GridLayer.extend({
 		this._levels = {};
 		this._tiles = {};
 		this._tileCache = {};
+		this._map._socket.sendMessage('commandvalues command=.uno:LanguageStatus');
 		this._map._socket.sendMessage('commandvalues command=.uno:ViewAnnotations');
 		var that = this;
 		$.contextMenu({
@@ -184,7 +190,7 @@ L.TileLayer = L.GridLayer.extend({
 						that.onAnnotationModify.call(that, options.$trigger.get(0).annotation);
 					}
 				},
-				reply: (this._docType !== 'text') ? undefined : {
+				reply: (this._docType !== 'text' && this._docType !== 'presentation') ? undefined : {
 					name: _('Reply'),
 					callback: function (key, options) {
 						that.onAnnotationReply.call(that, options.$trigger.get(0).annotation);
@@ -434,6 +440,9 @@ L.TileLayer = L.GridLayer.extend({
 		else if (textMsg.startsWith('unocommandresult:')) {
 			this._onUnoCommandResultMsg(textMsg);
 		}
+		else if (textMsg.startsWith('rulerupdate:')) {
+			this._onRulerUpdate(textMsg);
+		}
 		else if (textMsg.startsWith('contextmenu:')) {
 			this._onContextMenuMsg(textMsg);
 		}
@@ -455,6 +464,9 @@ L.TileLayer = L.GridLayer.extend({
 		else if (textMsg.startsWith('graphicviewselection:')) {
 			this._onGraphicViewSelectionMsg(textMsg);
 		}
+		else if (textMsg.startsWith('editor:')) {
+			this._updateEditor(textMsg);
+		}
 	},
 
 	toggleTileDebugMode: function() {
@@ -509,7 +521,9 @@ L.TileLayer = L.GridLayer.extend({
 
 	_onCellFormulaMsg: function (textMsg) {
 		var formula = textMsg.substring(13);
-		this._selectionTextContent = formula;
+		if (!this._map['wopi'].DisableCopy) {
+			this._selectionTextContent = formula;
+		}
 		this._map.fire('cellformula', {formula: formula});
 	},
 
@@ -524,8 +538,13 @@ L.TileLayer = L.GridLayer.extend({
 		var command = this._map._socket.parseServerCmd(textMsg);
 		var parser = document.createElement('a');
 		parser.href = this._map.options.server;
+
+		var wopiSrc = '';
+		if (map.options.wopiSrc != '') {
+			wopiSrc = '?WOPISrc=' + map.options.wopiSrc;
+		}
 		var url = this._map.options.webserver + '/' + this._map.options.urlPrefix + '/' +
-		    encodeURIComponent(this._map.options.doc) + '/' + command.jail + '/' + command.dir + '/' + command.name;
+		    encodeURIComponent(this._map.options.doc) + '/' + command.jail + '/' + command.dir + '/' + command.name + wopiSrc;
 
 		this._map.hideBusy();
 		if (command.id === 'print') {
@@ -624,13 +643,6 @@ L.TileLayer = L.GridLayer.extend({
 		if (!this._prevCellCursor) {
 			this._prevCellCursor = L.LatLngBounds.createDefault();
 		}
-		if (!this._cellCursorXY) {
-			this._cellCursorXY = new L.Point(-1, -1);
-		}
-		if (!this._prevCellCursorXY) {
-			this._prevCellCursorXY = new L.Point(-1, -1);
-		}
-
 		if (textMsg.match('EMPTY')) {
 			this._cellCursorTwips = new L.Bounds(new L.Point(0, 0), new L.Point(0, 0));
 			this._cellCursor = L.LatLngBounds.createDefault();
@@ -644,7 +656,6 @@ L.TileLayer = L.GridLayer.extend({
 			this._cellCursor = new L.LatLngBounds(
 							this._twipsToLatLng(topLeftTwips, this._map.getZoom()),
 							this._twipsToLatLng(bottomRightTwips, this._map.getZoom()));
-			this._cellCursorXY = new L.Point(parseInt(strTwips[4]), parseInt(strTwips[5]));
 		}
 
 		var horizontalDirection = 0;
@@ -708,6 +719,7 @@ L.TileLayer = L.GridLayer.extend({
 	},
 
 	_onInvalidateCursorMsg: function (textMsg) {
+		var docLayer = this._map._docLayer;
 		var strTwips = textMsg.match(/\d+/g);
 		var topLeftTwips = new L.Point(parseInt(strTwips[0]), parseInt(strTwips[1]));
 		var offset = new L.Point(parseInt(strTwips[2]), parseInt(strTwips[3]));
@@ -717,13 +729,33 @@ L.TileLayer = L.GridLayer.extend({
 						this._twipsToLatLng(bottomRightTwips, this._map.getZoom()));
 		this._visibleCursorOnLostFocus = this._visibleCursor;
 		this._isCursorOverlayVisible = true;
+		if ((docLayer._followEditor || docLayer._followUser) && this._map.lastActionByUser) {
+			this._map.fire('setFollowOff');
+		}
+		this._map.lastActionByUser = false;
 		this._onUpdateCursor();
 	},
 
+	_updateEditor: function(textMsg) {
+		textMsg = textMsg.substring('editor:'.length + 1);
+		var editorId = parseInt(textMsg);
+		var docLayer = this._map._docLayer;
+
+		docLayer._editorId = editorId;
+
+		if (docLayer._followEditor) {
+			docLayer._followThis = editorId;
+		}
+
+		if (this._map._viewInfo[editorId])
+			this._map.fire('updateEditorName', {username: this._map._viewInfo[editorId].username})
+	},
+
 	_onInvalidateViewCursorMsg: function (textMsg) {
 		textMsg = textMsg.substring('invalidateviewcursor:'.length + 1);
 		var obj = JSON.parse(textMsg);
 		var viewId = parseInt(obj.viewId);
+		var docLayer = this._map._docLayer;
 
 		// Ignore if viewid is same as ours or not in our db
 		if (viewId === this._viewId || !this._map._viewInfo[viewId]) {
@@ -747,6 +779,15 @@ L.TileLayer = L.GridLayer.extend({
 		}
 
 		this._onUpdateViewCursor(viewId);
+
+		if (docLayer._followThis === viewId && (docLayer._followEditor || docLayer._followUser)) {
+			if (this._map.getDocType() === 'text' || this._map.getDocType() === 'presentation') {
+				this.goToViewCursor(viewId);
+			}
+			else if (this._map.getDocType() === 'spreadsheet') {
+				this.goToCellViewCursor(viewId);
+			}
+		}
 	},
 
 	_onCellViewCursorMsg: function (textMsg) {
@@ -1007,7 +1048,6 @@ L.TileLayer = L.GridLayer.extend({
 		var index = textMsg.indexOf('=');
 		var commandName = index !== -1 ? textMsg.substr(0, index) : '';
 		var state = index !== -1 ? textMsg.substr(index + 1) : '';
-
 		this._map.fire('commandstatechanged', {commandName : commandName, state : state});
 	},
 
@@ -1027,6 +1067,13 @@ L.TileLayer = L.GridLayer.extend({
 
 	},
 
+	_onRulerUpdate: function (textMsg) {
+		textMsg = textMsg.substring(13);
+		var obj = JSON.parse(textMsg);
+
+		this._map.fire('rulerupdate', obj);
+	},
+
 	_onContextMenuMsg: function (textMsg) {
 		textMsg = textMsg.substring(13);
 		var obj = JSON.parse(textMsg);
@@ -1183,8 +1230,8 @@ L.TileLayer = L.GridLayer.extend({
 			});
 		}
 		else if (tile) {
-			if (command.hash != undefined) {
-				tile.oldhash = command.hash;
+			if (command.wireId != undefined) {
+				tile.oldWireId = command.wireId;
 			}
 			if (this._tiles[key]._invalidCount > 0) {
 				this._tiles[key]._invalidCount -= 1;
@@ -1321,6 +1368,7 @@ L.TileLayer = L.GridLayer.extend({
 	// Update cursor layer (blinking cursor).
 	_onUpdateCursor: function (e) {
 		var cursorPos = this._visibleCursor.getNorthWest();
+		var docLayer = this._map._docLayer;
 
 		if (!e && !this._map.getBounds().contains(this._visibleCursor) && this._isCursorVisible) {
 			var center = this._map.project(cursorPos);
@@ -1329,7 +1377,8 @@ L.TileLayer = L.GridLayer.extend({
 			center.y = Math.round(center.y < 0 ? 0 : center.y);
 
 			if (!(this._selectionHandles.start && this._selectionHandles.start.isDragged) &&
-			    !(this._selectionHandles.end && this._selectionHandles.end.isDragged)) {
+			    !(this._selectionHandles.end && this._selectionHandles.end.isDragged) &&
+			    !(docLayer._followEditor || docLayer._followUser)) {
 				this._map.fire('scrollto', {x: center.x, y: center.y});
 			}
 		}
@@ -1596,7 +1645,7 @@ L.TileLayer = L.GridLayer.extend({
 	_onUpdateCellCursor: function (horizontalDirection, verticalDirection, onPgUpDn) {
 		if (this._cellCursor && !this._isEmptyRectangle(this._cellCursor)) {
 			var mapBounds = this._map.getBounds();
-			if (!mapBounds.contains(this._cellCursor) && !this._cellCursorXY.equals(this._prevCellCursorXY)) {
+			if (!mapBounds.contains(this._cellCursor)) {
 				var scrollX = 0, scrollY = 0;
 				if (onPgUpDn) {
 					var mapHalfHeight = (mapBounds.getNorth() - mapBounds.getSouth()) / 2;
@@ -1634,7 +1683,6 @@ L.TileLayer = L.GridLayer.extend({
 					center.y = Math.round(center.y < 0 ? 0 : center.y);
 					this._map.fire('scrollto', {x: center.x, y: center.y});
 				}
-				this._prevCellCursorXY = this._cellCursorXY;
 			}
 
 			if (onPgUpDn) {
