@@ -100,7 +100,7 @@ L.WriterTileLayer = L.TileLayer.extend({
 		var visibleArea = new L.Bounds(visibleTopLeft, visibleBottomRight);
 		var tilePositionsX = '';
 		var tilePositionsY = '';
-		var oldHashes = '';
+		var oldWireIds = '';
 		var needsNewTiles = false;
 		for (var key in this._tiles) {
 			var coords = this._tiles[key].coords;
@@ -123,14 +123,14 @@ L.WriterTileLayer = L.TileLayer.extend({
 						tilePositionsY += ',';
 					}
 					tilePositionsY += tileTopLeft.y;
-					if (oldHashes !== '') {
-						oldHashes += ',';
+					if (oldWireIds !== '') {
+						oldWireIds += ',';
 					}
-					if (this._tiles[key].oldhash === undefined) {
-						oldHashes += '0';
+					if (this._tiles[key].oldWireId === undefined) {
+						oldWireIds += '0';
 					}
 					else {
-						oldHashes += this._tiles[key].oldhash;
+						oldWireIds += this._tiles[key].oldWireId;
 					}
 					needsNewTiles = true;
 					if (this._debug) {
@@ -158,7 +158,7 @@ L.WriterTileLayer = L.TileLayer.extend({
 				'tileposy=' + tilePositionsY + ' ' +
 				'tilewidth=' + this._tileWidthTwips + ' ' +
 				'tileheight=' + this._tileHeightTwips + ' ' +
-				'oldhash=' + oldHashes;
+				'oldwid=' + oldWireIds;
 
 			this._map._socket.sendMessage(message, '');
 
@@ -230,11 +230,11 @@ L.WriterTileLayer = L.TileLayer.extend({
 		this._update();
 	},
 
-	_updateMaxBounds: function (sizeChanged, extraSize) {
+	_updateMaxBounds: function (sizeChanged, extraSize, zoom) {
 		if (!extraSize) {
 			extraSize = this._annotations && this._annotations._items.length > 0 ?
 				this._annotations.options.extraSize : null;
 		}
-		L.GridLayer.prototype._updateMaxBounds.call(this, sizeChanged, extraSize, {panInside: false});
+		L.GridLayer.prototype._updateMaxBounds.call(this, sizeChanged, extraSize, {panInside: false}, zoom);
 	}
 });
