@@ -14,6 +14,7 @@
 #include <algorithm>
 #include <cassert>
 #include <fstream>
+#include <iconv.h>
 #include <string>
 
 #include <Poco/DateTime.h>
@@ -33,19 +34,19 @@
 #include <Poco/Timestamp.h>
 
 // For residual Poco SSL usage.
-#include <Poco/Net/Context.h>
-#include <Poco/Net/SSLManager.h>
 #include <Poco/Net/AcceptCertificateHandler.h>
+#include <Poco/Net/Context.h>
 #include <Poco/Net/KeyConsoleHandler.h>
+#include <Poco/Net/SSLManager.h>
 
 #include "Auth.hpp"
 #include "Common.hpp"
 #include "Exceptions.hpp"
-#include "common/FileUtil.hpp"
 #include "LOOLWSD.hpp"
 #include "Log.hpp"
 #include "Unit.hpp"
 #include "Util.hpp"
+#include "common/FileUtil.hpp"
 
 bool StorageBase::FilesystemEnabled;
 bool StorageBase::WopiEnabled;
@@ -195,6 +196,21 @@ std::unique_ptr<StorageBase> StorageBase::create(const Poco::URI& uri, const std
         {
             return std::unique_ptr<StorageBase>(new LocalStorage(uri, jailRoot, jailPath));
         }
+        else
+        {
+            // guard against attempts to escape
+            Poco::URI normalizedUri(uri);
+            normalizedUri.normalize();
+
+            std::vector<std::string> pathSegments;
+            normalizedUri.getPathSegments(pathSegments);
+
+            if (pathSegments.size() == 4 && pathSegments[0] == "tmp" && pathSegments[1] == "convert-to")
+            {
+                LOG_INF("Public URI [" << normalizedUri.toString() << "] is actually a convert-to tempfile.");
+                return std::unique_ptr<StorageBase>(new LocalStorage(normalizedUri, jailRoot, jailPath));
+            }
+        }
 
         LOG_ERR("Local Storage is disabled by default. Enable in the config file or on the command-line to enable.");
     }
@@ -231,7 +247,7 @@ std::unique_ptr<LocalStorage::LocalFileInfo> LocalStorage::getLocalFileInfo()
     return std::unique_ptr<LocalStorage::LocalFileInfo>(new LocalFileInfo({"localhost" + std::to_string(LastLocalStorageId), "Local Host #" + std::to_string(LastLocalStorageId++)}));
 }
 
-std::string LocalStorage::loadStorageFileToLocal(const std::string& /*accessToken*/)
+std::string LocalStorage::loadStorageFileToLocal(const Authorization& /*auth*/)
 {
     // /chroot/jailId/user/doc/childId/file.ext
     const auto filename = Poco::Path(_uri.getPath()).getFileName();
@@ -282,16 +298,22 @@ std::string LocalStorage::loadStorageFileToLocal(const std::string& /*accessToke
 #endif
 }
 
-StorageBase::SaveResult LocalStorage::saveLocalFileToStorage(const std::string& /*accessToken*/)
+StorageBase::SaveResult LocalStorage::saveLocalFileToStorage(const Authorization& /*auth*/, const std::string& /*saveAsPath*/, const std::string& /*saveAsFilename*/)
 {
     try
     {
+        LOG_TRC("Saving local file to local file storage " << _isCopy << " for " << _jailedFilePath);
         // Copy the file back.
         if (_isCopy && Poco::File(_jailedFilePath).exists())
         {
             LOG_INF("Copying " << _jailedFilePath << " to " << _uri.getPath());
             Poco::File(_jailedFilePath).copyTo(_uri.getPath());
         }
+
+        // update its fileinfo object. This is used later to check if someone else changed the
+        // document while we are/were editing it
+        _fileInfo._modifiedTime = Poco::File(_uri.getPath()).getLastModified();
+        Log::trace() << "New FileInfo modified time in storage " << _fileInfo._modifiedTime << Log::end;
     }
     catch (const Poco::Exception& exc)
     {
@@ -300,7 +322,7 @@ StorageBase::SaveResult LocalStorage::saveLocalFileToStorage(const std::string&
         return StorageBase::SaveResult::FAILED;
     }
 
-    return StorageBase::SaveResult::OK;
+    return StorageBase::SaveResult(StorageBase::SaveResult::OK);
 }
 
 namespace {
@@ -347,6 +369,27 @@ int getLevenshteinDist(const std::string& string1, const std::string& string2) {
     return matrix[string1.size()][string2.size()];
 }
 
+// Gets value for `key` directly from the given JSON in `object`
+template <typename T>
+T getJSONValue(const Poco::JSON::Object::Ptr &object, const std::string& key)
+{
+    T value = T();
+    try
+    {
+        const Poco::Dynamic::Var valueVar = object->get(key);
+        value = valueVar.convert<T>();
+    }
+    catch (const Poco::Exception& exc)
+    {
+        LOG_ERR("getJSONValue: " << exc.displayText() <<
+                (exc.nested() ? " (" + exc.nested()->displayText() + ")" : ""));
+    }
+
+    return value;
+}
+
+// Function that searches `object` for `key` and warns if there are minor mis-spellings involved
+// Upon successfull search, fills `value` with value found in object.
 template <typename T>
 void getWOPIValue(const Poco::JSON::Object::Ptr &object, const std::string& key, T& value)
 {
@@ -371,41 +414,31 @@ void getWOPIValue(const Poco::JSON::Object::Ptr &object, const std::string& key,
             return;
         }
 
-        try
-        {
-            const Poco::Dynamic::Var valueVar = object->get(userInput);
-            value = valueVar.convert<T>();
-        }
-        catch (const Poco::Exception& exc)
-        {
-            LOG_ERR("getWOPIValue: " << exc.displayText() <<
-                    (exc.nested() ? " (" + exc.nested()->displayText() + ")" : ""));
-        }
-
+        value = getJSONValue<T>(object, userInput);
         return;
     }
 
     LOG_WRN("Missing JSON property [" << key << "]");
 }
 
-void setQueryParameter(Poco::URI& uriObject, const std::string& key, const std::string& value)
+// Parse the json string and fill the Poco::JSON object
+// Returns true if parsing successful otherwise false
+bool parseJSON(const std::string& json, Poco::JSON::Object::Ptr& object)
 {
-    Poco::URI::QueryParameters queryParams = uriObject.getQueryParameters();
-    for (auto& param: queryParams)
+    bool success = false;
+    const auto index = json.find_first_of('{');
+    if (index != std::string::npos)
     {
-        if (param.first == key)
-        {
-            param.second = value;
-            uriObject.setQueryParameters(queryParams);
-            return;
-        }
+        const std::string stringJSON = json.substr(index);
+        Poco::JSON::Parser parser;
+        const auto result = parser.parse(stringJSON);
+        object = result.extract<Poco::JSON::Object::Ptr>();
+        success = true;
     }
 
-    // it did not exist yet
-    uriObject.addQueryParameter(key, value);
+    return success;
 }
 
-
 void addStorageDebugCookie(Poco::Net::HTTPRequest& request)
 {
     (void) request;
@@ -424,13 +457,32 @@ void addStorageDebugCookie(Poco::Net::HTTPRequest& request)
 #endif
 }
 
+Poco::Timestamp iso8601ToTimestamp(const std::string& iso8601Time)
+{
+    Poco::Timestamp timestamp = Poco::Timestamp::fromEpochTime(0);
+    try
+    {
+        int timeZoneDifferential;
+        Poco::DateTime dateTime;
+        Poco::DateTimeParser::parse(Poco::DateTimeFormat::ISO8601_FRAC_FORMAT, iso8601Time, dateTime, timeZoneDifferential);
+        timestamp = dateTime.timestamp();
+    }
+    catch (const Poco::SyntaxException& exc)
+    {
+        LOG_WRN("Time [" << iso8601Time << "] is in invalid format: " << exc.displayText() <<
+                (exc.nested() ? " (" + exc.nested()->displayText() + ")" : ""));
+    }
+
+    return timestamp;
+}
+
 } // anonymous namespace
 
-std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const std::string& accessToken)
+std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const Authorization& auth)
 {
     // update the access_token to the one matching to the session
     Poco::URI uriObject(_uri);
-    setQueryParameter(uriObject, "access_token", accessToken);
+    auth.authorizeURI(uriObject);
 
     LOG_DBG("Getting info for wopi uri [" << uriObject.toString() << "].");
 
@@ -443,6 +495,7 @@ std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const st
 
         Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_GET, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
         request.set("User-Agent", WOPI_AGENT_STRING);
+        auth.authorizeRequest(request);
         addStorageDebugCookie(request);
         psession->sendRequest(request);
 
@@ -483,6 +536,8 @@ std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const st
     std::string ownerId;
     std::string userId;
     std::string userName;
+    std::string userExtraInfo;
+    std::string watermarkText;
     bool canWrite = false;
     bool enableOwnerTermination = false;
     std::string postMessageOrigin;
@@ -492,21 +547,21 @@ std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const st
     bool disablePrint = false;
     bool disableExport = false;
     bool disableCopy = false;
+    bool disableInactiveMessages = false;
     std::string lastModifiedTime;
+    bool userCanNotWriteRelative = true;
 
     LOG_DBG("WOPI::CheckFileInfo returned: " << resMsg << ". Call duration: " << callDuration.count() << "s");
-    const auto index = resMsg.find_first_of('{');
-    if (index != std::string::npos)
+    Poco::JSON::Object::Ptr object;
+    if (parseJSON(resMsg, object))
     {
-        const std::string stringJSON = resMsg.substr(index);
-        Poco::JSON::Parser parser;
-        const auto result = parser.parse(stringJSON);
-        const auto& object = result.extract<Poco::JSON::Object::Ptr>();
         getWOPIValue(object, "BaseFileName", filename);
         getWOPIValue(object, "Size", size);
         getWOPIValue(object, "OwnerId", ownerId);
         getWOPIValue(object, "UserId", userId);
         getWOPIValue(object, "UserFriendlyName", userName);
+        getWOPIValue(object, "UserExtraInfo", userExtraInfo);
+        getWOPIValue(object, "WatermarkText", watermarkText);
         getWOPIValue(object, "UserCanWrite", canWrite);
         getWOPIValue(object, "PostMessageOrigin", postMessageOrigin);
         getWOPIValue(object, "HidePrintOption", hidePrintOption);
@@ -516,7 +571,9 @@ std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const st
         getWOPIValue(object, "DisablePrint", disablePrint);
         getWOPIValue(object, "DisableExport", disableExport);
         getWOPIValue(object, "DisableCopy", disableCopy);
+        getWOPIValue(object, "DisableInactiveMessages", disableInactiveMessages);
         getWOPIValue(object, "LastModifiedTime", lastModifiedTime);
+        getWOPIValue(object, "UserCanNotWriteRelative", userCanNotWriteRelative);
     }
     else
     {
@@ -524,41 +581,21 @@ std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const st
         throw UnauthorizedRequestException("Access denied. WOPI::CheckFileInfo failed on: " + uriObject.toString());
     }
 
-    Poco::Timestamp modifiedTime = Poco::Timestamp::fromEpochTime(0);
-    if (lastModifiedTime != "")
-    {
-        Poco::DateTime dateTime;
-        int timeZoneDifferential;
-        bool valid = false;
-        try
-        {
-            Poco::DateTimeParser::parse(Poco::DateTimeFormat::ISO8601_FRAC_FORMAT, lastModifiedTime, dateTime, timeZoneDifferential);
-            valid = true;
-        }
-        catch (const Poco::SyntaxException& exc)
-        {
-            LOG_WRN("LastModifiedTime property [" << lastModifiedTime << "] was invalid format: " << exc.displayText() <<
-                    (exc.nested() ? " (" + exc.nested()->displayText() + ")" : ""));
-        }
-        if (valid)
-        {
-            modifiedTime = dateTime.timestamp();
-        }
-    }
-
+    const Poco::Timestamp modifiedTime = iso8601ToTimestamp(lastModifiedTime);
     _fileInfo = FileInfo({filename, ownerId, modifiedTime, size});
 
-    return std::unique_ptr<WopiStorage::WOPIFileInfo>(new WOPIFileInfo({userId, userName, canWrite, postMessageOrigin, hidePrintOption, hideSaveOption, hideExportOption, enableOwnerTermination, disablePrint, disableExport, disableCopy, callDuration}));
+    return std::unique_ptr<WopiStorage::WOPIFileInfo>(new WOPIFileInfo({userId, userName, userExtraInfo, watermarkText, canWrite, postMessageOrigin, hidePrintOption, hideSaveOption, hideExportOption, enableOwnerTermination, disablePrint, disableExport, disableCopy, disableInactiveMessages, userCanNotWriteRelative, callDuration}));
 }
 
 /// uri format: http://server/<...>/wopi*/files/<id>/content
-std::string WopiStorage::loadStorageFileToLocal(const std::string& accessToken)
+std::string WopiStorage::loadStorageFileToLocal(const Authorization& auth)
 {
     // WOPI URI to download files ends in '/contents'.
     // Add it here to get the payload instead of file info.
     Poco::URI uriObject(_uri);
     uriObject.setPath(uriObject.getPath() + "/contents");
-    setQueryParameter(uriObject, "access_token", accessToken);
+    auth.authorizeURI(uriObject);
+
     LOG_DBG("Wopi requesting: " << uriObject.toString());
 
     const auto startTime = std::chrono::steady_clock::now();
@@ -568,6 +605,7 @@ std::string WopiStorage::loadStorageFileToLocal(const std::string& accessToken)
 
         Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_GET, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
         request.set("User-Agent", WOPI_AGENT_STRING);
+        auth.authorizeRequest(request);
         addStorageDebugCookie(request);
         psession->sendRequest(request);
 
@@ -600,7 +638,6 @@ std::string WopiStorage::loadStorageFileToLocal(const std::string& accessToken)
             std::copy(std::istreambuf_iterator<char>(rs),
                       std::istreambuf_iterator<char>(),
                       std::ostreambuf_iterator<char>(ofs));
-            ofs.close();
             LOG_INF("WOPI::GetFile downloaded " << getFileSize(_jailedFilePath) << " bytes from [" << uriObject.toString() <<
                     "] -> " << _jailedFilePath << " in " << diff.count() << "s");
 
@@ -619,74 +656,178 @@ std::string WopiStorage::loadStorageFileToLocal(const std::string& accessToken)
     return "";
 }
 
-StorageBase::SaveResult WopiStorage::saveLocalFileToStorage(const std::string& accessToken)
+StorageBase::SaveResult WopiStorage::saveLocalFileToStorage(const Authorization& auth, const std::string& saveAsPath, const std::string& saveAsFilename)
 {
     // TODO: Check if this URI has write permission (canWrite = true)
-    const auto size = getFileSize(_jailedFilePath);
+
+    const bool isSaveAs = !saveAsPath.empty() && !saveAsFilename.empty();
+    const std::string filePath(isSaveAs? saveAsPath: _jailedFilePath);
+
+    const auto size = getFileSize(filePath);
 
     Poco::URI uriObject(_uri);
-    uriObject.setPath(uriObject.getPath() + "/contents");
-    setQueryParameter(uriObject, "access_token", accessToken);
+    uriObject.setPath(isSaveAs? uriObject.getPath(): uriObject.getPath() + "/contents");
+    auth.authorizeURI(uriObject);
 
-    LOG_INF("Uploading URI via WOPI [" << uriObject.toString() << "] from [" << _jailedFilePath + "].");
+    LOG_INF("Uploading URI via WOPI [" << uriObject.toString() << "] from [" << filePath + "].");
 
     std::ostringstream oss;
-    StorageBase::SaveResult saveResult = StorageBase::SaveResult::FAILED;
+    StorageBase::SaveResult saveResult(StorageBase::SaveResult::FAILED);
     try
     {
         std::unique_ptr<Poco::Net::HTTPClientSession> psession(getHTTPClientSession(uriObject));
 
         Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_POST, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
-        request.set("X-WOPI-Override", "PUT");
+        request.set("User-Agent", WOPI_AGENT_STRING);
+        auth.authorizeRequest(request);
+
+        if (!isSaveAs)
+        {
+            // normal save
+            request.set("X-WOPI-Override", "PUT");
+            request.set("X-LOOL-WOPI-IsModifiedByUser", _isUserModified? "true": "false");
+            request.set("X-LOOL-WOPI-IsAutosave", _isAutosave? "true": "false");
+
+            if (!_forceSave)
+            {
+                // Request WOPI host to not overwrite if timestamps mismatch
+                request.set("X-LOOL-WOPI-Timestamp",
+                            Poco::DateTimeFormatter::format(Poco::DateTime(_fileInfo._modifiedTime),
+                                                            Poco::DateTimeFormat::ISO8601_FRAC_FORMAT));
+            }
+        }
+        else
+        {
+            // save as
+            request.set("X-WOPI-Override", "PUT_RELATIVE");
+
+            // the suggested target has to be in UTF-7; default to extension
+            // only when the conversion fails
+            std::string suggestedTarget = "." + Poco::Path(saveAsFilename).getExtension();
+
+            iconv_t cd = iconv_open("UTF-7", "UTF-8");
+            if (cd == (iconv_t) -1)
+                LOG_ERR("Failed to initialize iconv for UTF-7 conversion, using '" << suggestedTarget << "'.");
+            else
+            {
+                std::vector<char> input(saveAsFilename.begin(), saveAsFilename.end());
+                std::vector<char> buffer(8 * saveAsFilename.size());
+
+                char* in = &input[0];
+                size_t in_left = input.size();
+                char* out = &buffer[0];
+                size_t out_left = buffer.size();
+
+                if (iconv(cd, &in, &in_left, &out, &out_left) == (size_t) -1)
+                    LOG_ERR("Failed to convert '" << saveAsFilename << "' to UTF-7, using '" << suggestedTarget << "'.");
+                else
+                {
+                    // conversion succeeded
+                    suggestedTarget = std::string(&buffer[0], buffer.size() - out_left);
+                    LOG_TRC("Converted '" << saveAsFilename << "' to UTF-7 as '" << suggestedTarget << "'.");
+                }
+            }
+
+            request.set("X-WOPI-SuggestedTarget", suggestedTarget);
+
+            request.set("X-WOPI-Size", std::to_string(size));
+        }
+
         request.setContentType("application/octet-stream");
         request.setContentLength(size);
         addStorageDebugCookie(request);
         std::ostream& os = psession->sendRequest(request);
-        std::ifstream ifs(_jailedFilePath);
+
+        std::ifstream ifs(filePath);
         Poco::StreamCopier::copyStream(ifs, os);
 
         Poco::Net::HTTPResponse response;
         std::istream& rs = psession->receiveResponse(response);
         Poco::StreamCopier::copyStream(rs, oss);
-        LOG_INF("WOPI::PutFile response: " << oss.str());
-        LOG_INF("WOPI::PutFile uploaded " << size << " bytes from [" << _jailedFilePath <<
+
+        std::string wopiLog(isSaveAs? "WOPI::PutRelativeFile": "WOPI::PutFile");
+        LOG_INF(wopiLog << " response: " << oss.str());
+        LOG_INF(wopiLog << " uploaded " << size << " bytes from [" << filePath <<
                 "] -> [" << uriObject.toString() << "]: " <<
                 response.getStatus() << " " << response.getReason());
 
         if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_OK)
         {
-            saveResult = StorageBase::SaveResult::OK;
+            saveResult.setResult(StorageBase::SaveResult::OK);
+            Poco::JSON::Object::Ptr object;
+            if (parseJSON(oss.str(), object))
+            {
+                const std::string lastModifiedTime = getJSONValue<std::string>(object, "LastModifiedTime");
+                LOG_TRC(wopiLog << " returns LastModifiedTime [" << lastModifiedTime << "].");
+                _fileInfo._modifiedTime = iso8601ToTimestamp(lastModifiedTime);
+
+                if (isSaveAs)
+                {
+                    const std::string name = getJSONValue<std::string>(object, "Name");
+                    LOG_TRC(wopiLog << " returns Name [" << name << "].");
+
+                    const std::string url = getJSONValue<std::string>(object, "Url");
+                    LOG_TRC(wopiLog << " returns Url [" << url << "].");
+
+                    saveResult.setSaveAsResult(name, url);
+                }
+
+                // Reset the force save flag now, if any, since we are done saving
+                // Next saves shouldn't be saved forcefully unless commanded
+                _forceSave = false;
+            }
+            else
+            {
+                LOG_WRN("Invalid or missing JSON in " << wopiLog << " HTTP_OK response");
+            }
         }
         else if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_REQUESTENTITYTOOLARGE)
         {
-            saveResult = StorageBase::SaveResult::DISKFULL;
+            saveResult.setResult(StorageBase::SaveResult::DISKFULL);
         }
         else if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_UNAUTHORIZED)
         {
-            saveResult = StorageBase::SaveResult::UNAUTHORIZED;
+            saveResult.setResult(StorageBase::SaveResult::UNAUTHORIZED);
+        }
+        else if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_CONFLICT)
+        {
+            saveResult.setResult(StorageBase::SaveResult::CONFLICT);
+            Poco::JSON::Object::Ptr object;
+            if (parseJSON(oss.str(), object))
+            {
+                const unsigned loolStatusCode = getJSONValue<unsigned>(object, "LOOLStatusCode");
+                if (loolStatusCode == static_cast<unsigned>(LOOLStatusCode::DOC_CHANGED))
+                {
+                    saveResult.setResult(StorageBase::SaveResult::DOC_CHANGED);
+                }
+            }
+            else
+            {
+                LOG_WRN("Invalid or missing JSON in " << wopiLog << " HTTP_CONFLICT response");
+            }
         }
     }
     catch(const Poco::Exception& pexc)
     {
         LOG_ERR("Cannot save file to WOPI storage uri [" + uriObject.toString() + "]. Error: " << pexc.displayText() <<
                 (pexc.nested() ? " (" + pexc.nested()->displayText() + ")" : ""));
-        saveResult = StorageBase::SaveResult::FAILED;
+        saveResult.setResult(StorageBase::SaveResult::FAILED);
     }
 
     return saveResult;
 }
 
-std::string WebDAVStorage::loadStorageFileToLocal(const std::string& /*accessToken*/)
+std::string WebDAVStorage::loadStorageFileToLocal(const Authorization& /*auth*/)
 {
     // TODO: implement webdav GET.
     _isLoaded = true;
     return _uri.toString();
 }
 
-StorageBase::SaveResult WebDAVStorage::saveLocalFileToStorage(const std::string& /*accessToken*/)
+StorageBase::SaveResult WebDAVStorage::saveLocalFileToStorage(const Authorization& /*auth*/, const std::string& /*saveAsPath*/, const std::string& /*saveAsFilename*/)
 {
     // TODO: implement webdav PUT.
-    return StorageBase::SaveResult::OK;
+    return StorageBase::SaveResult(StorageBase::SaveResult::OK);
 }
 
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
